==> v1.0.3.tar.gz: activiti-api-doclet/src/main/java/com/inteligr8/activiti/doclet/MarkdownWriter.java <==
/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.activiti.doclet;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Name;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.TypeMirror;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sun.source.doctree.DocCommentTree;
import com.sun.source.doctree.DocTree;
import com.sun.source.util.DocTrees;

import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;
import jdk.javadoc.doclet.DocletEnvironment;

/**
 * @author brian@inteligr8.com
 */
class MarkdownWriter {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final Set<String> bpmnErrorSet = new HashSet<>(Arrays.asList(
			"BpmnError".toLowerCase(),
			"org.activiti.engine.delegate.BpmnError".toLowerCase()));
	
	private final Pattern tagPattern = Pattern.compile("@([^ ]+) ?(.*)$");
	private final Pattern namedCommentPattern = Pattern.compile("([^ ]+) ?(.*)$");
	private final Map<Pattern, String> sanitizations = new FluentMap<>(new LinkedHashMap<Pattern, String>())
			.with(Pattern.compile("^[ ]+", Pattern.MULTILINE), "")
			.with(Pattern.compile("\r\n", Pattern.MULTILINE), "\n")
			.with(Pattern.compile("([^\n])\n([^\n])", Pattern.MULTILINE), "$1 $2")
			.with(Pattern.compile("([^\\.\\?!])[ ]{2}", Pattern.MULTILINE), "$1 ")
			.with(Pattern.compile("\\{ ?(@[^}]+) \\}"), "{$1}")
			.with(Pattern.compile("\\{@link ([^}]+)\\}"), "[$1]($1)")
			.with(Pattern.compile("([A-Za-z0-9\\-\\._]+@[A-Za-z0-9\\-\\._]+)"), "[$1](mailto:$1)");
	private final Pattern seeRefPattern = Pattern.compile("^([A-Za-z0-9_\\-]+)\\.([A-Za-z0-9_]+)$");
	
	private final DocletEnvironment docenv;
	private final File outputDirectory;
	private final Configuration fmconfig;
	
	private String title = "API Documentation";
	private String apiName;
	
	public MarkdownWriter(DocletEnvironment docenv) throws IOException {
		this.docenv = docenv;
		this.outputDirectory = new File(".");
		this.fmconfig = this.getFreemarkerConfiguration();
	}
	
	private Configuration getFreemarkerConfiguration() throws IOException {
		Configuration fmconfig = new Configuration(Configuration.VERSION_2_3_31);
		fmconfig.setDefaultEncoding("utf-8");
		fmconfig.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);
		fmconfig.setLogTemplateExceptions(false);
		fmconfig.setWrapUncheckedExceptions(true);
		fmconfig.setFallbackOnNullLoopVariable(false);
		fmconfig.setClassLoaderForTemplateLoading(this.getClass().getClassLoader(), "/templates");
		return fmconfig;
	}
	
	public void setTitle(String title) {
		this.title = title;
	}
	
	public void setApiName(String apiName) {
		this.apiName = apiName;
	}
	
	public void write(List<ActivitiApiBeanDoc> beandocs) throws IOException, TemplateException {
		IndexFreemarkerModel model = this.buildModel(beandocs);
		
		this.outputDirectory.mkdirs();
		this.writeIndexFile(model);
		for (BeanFreemarkerModel beanModel : model.getBeans())
			this.writeBeanFile(beanModel);
	}
	
	private IndexFreemarkerModel buildModel(List<ActivitiApiBeanDoc> beandocs) throws IOException, TemplateException {
		DocTrees docs = this.docenv.getDocTrees();

		this.logger.debug("Building documentation index model");
		
		IndexFreemarkerModel indexModel = new IndexFreemarkerModel();
		indexModel.setTitle(this.title);
		indexModel.setApiName(this.apiName);

		for (ActivitiApiBeanDoc beandoc : beandocs) {
			DocCommentTree commentTree = docs.getDocCommentTree(beandoc.getClassElement());

			this.logger.debug("Building documentation bean model: {}", beandoc.getBeanId());
	
			BeanFreemarkerModel beanModel = new BeanFreemarkerModel();
			beanModel.setTitle(this.title);
			beanModel.setApiName(this.apiName);
			beanModel.setBeanId(beandoc.getBeanId());
			if (beandoc.isDelegate())
				beanModel.getDelegateRoles().add("ServiceTask");
			if (beandoc.isExecutionListener())
				beanModel.getDelegateRoles().add("ExecutionListener");
			if (beandoc.isTaskListener())
				beanModel.getDelegateRoles().add("TaskListener");
			if (commentTree != null) {
				this.logger.debug("Found documentation for bean: {}", beandoc.getBeanId());
				
				this.buildDocumentation(commentTree, beanModel);
				for (DocTree tag : commentTree.getBlockTags()) {
					this.buildTagModel(tag, beandoc.getBeanId(),
							null, null, null, beanModel);
				}
			}

			Map<Name, MethodFreemarkerModel> methodOverloads = new LinkedHashMap<>();
			
			for (ExecutableElement methodElement : beandoc.getSortedMethodElements()) {
				String logId = beandoc.getBeanId() + "." + methodElement.toString();
				DocCommentTree methodCommentTree = docs.getDocCommentTree(methodElement);

				this.logger.debug("Building documentation method model: {}", logId);
				
				MethodFreemarkerModel methodModel = methodOverloads.get(methodElement.getSimpleName());
				if (methodModel == null) {
					methodModel = new MethodFreemarkerModel();
					methodModel.setMethodName(methodElement.getSimpleName().toString());
					
					methodOverloads.put(methodElement.getSimpleName(), methodModel);
					beanModel.getMethods().add(methodModel);
				}
				
				Map<String, String> paramComments = new HashMap<>();
				Map<String, String> throwComments = new LinkedHashMap<>();
				Map<String, String> errorComments = new LinkedHashMap<>();
				
				MethodSignatureFreemarkerModel sigModel = new MethodSignatureFreemarkerModel();
				if (methodCommentTree != null) {
					this.logger.debug("Found documentation for '{}'", logId);

					this.buildDocumentation(methodCommentTree, sigModel);
					if (methodModel.getDocFirstSentence() == null)
						methodModel.setDocFirstSentence(sigModel.getDocFirstSentence());
					for (DocTree tag : methodCommentTree.getBlockTags()) {
						this.buildTagModel(tag, logId,
								paramComments, throwComments, errorComments, sigModel);
					}
				}
				
				this.buildMethodArguments(methodElement, logId, paramComments, sigModel);
				this.buildMethodThrows(methodElement, logId, throwComments, !errorComments.isEmpty(), sigModel);
				this.buildMethodErrors(methodElement, logId, errorComments, sigModel);
				this.buildMethodReturns(methodElement, logId, sigModel);
				this.buildMethodSignature(methodElement, sigModel);

				if (!paramComments.isEmpty())
					this.logger.warn("'{}' has documented parameters {} that don't exist; ignoring ...", logId, paramComments.keySet());
				
				methodModel.getSignatures().add(sigModel);
			}
			
			indexModel.getBeans().add(beanModel);
		}
		
		return indexModel;
	}
	
	private void buildDocumentation(DocCommentTree commentTree, JavadocDocumentable docuable) {
		if (commentTree.getFirstSentence() != null)
			docuable.setDocFirstSentence(this.sanitize(commentTree.getFirstSentence()));
		if (commentTree.getBody() != null)
			docuable.setDocBody(this.sanitize(commentTree.getFullBody()));
	}
	
	private void buildTagModel(DocTree tag, String logId,
			Map<String, String> paramComments,
			Map<String, String> throwComments,
			Map<String, String> errorComments,
			JavadocTaggable sigModel) {
		Matcher matcher = this.tagPattern.matcher(tag.toString());
		if (!matcher.find()) {
			this.logger.warn("A tag for '{}' is improperly formatted", logId);
			return;
		}
		
		String tagName = matcher.group(1);
		String tagComment = matcher.group(2).trim();

		this.logger.trace("Found tag '{}' for '{}'", tagName, logId);

		switch (tagName) {
			case "param":
				if (paramComments == null)
					break;
				
				matcher = this.namedCommentPattern.matcher(tagComment);
				if (matcher.find())
					paramComments.put(matcher.group(1), this.sanitize(matcher.group(2).trim()));
				else this.logger.warn("A @param for the {} bean and {} method is improperly formatted", logId);
				return;
			case "throws":
				if (throwComments == null)
					break;
				
				matcher = this.namedCommentPattern.matcher(tagComment);
				if (matcher.find()) {
					if (this.bpmnErrorSet.contains(matcher.group(1).toLowerCase())) {
						Matcher ecmatcher = this.namedCommentPattern.matcher(matcher.group(2).trim());
						if (ecmatcher.find()) {
							errorComments.put(ecmatcher.group(1), this.sanitize(ecmatcher.group(2).trim()));
						} else {
							this.logger.warn("A @throws for the {} bean and {} method is improperly formatted", logId);
						}
					} else {
						throwComments.put(matcher.group(1), this.sanitize(matcher.group(2).trim()));
					}
				}
				else this.logger.warn("A @throws for the {} bean and {} method is improperly formatted", logId);
				return;
			case "see":
				matcher = this.seeRefPattern.matcher(tagComment);
				if (matcher.find()) {
					ReferenceFreemarkerModel model = new ReferenceFreemarkerModel();
					model.setBeanId(matcher.group(1));
					model.setMethodName(matcher.group(2));
					sigModel.getSeeRefs().add(model);
					return;
				}
			default:
		}

		this.putAdd(sigModel.getTags(), tagName, this.sanitize(tagComment));
	}
	
	private void buildMethodArguments(ExecutableElement methodElement, String logId,
			Map<String, String> paramComments, MethodSignatureFreemarkerModel sigModel) {
		for (VariableElement varElement : methodElement.getParameters()) {
			this.logger.trace("Found parameter '{}' for '{}'", varElement.getSimpleName(), logId);
			
			ParamFreemarkerModel paramModel = new ParamFreemarkerModel();
			paramModel.setName(varElement.getSimpleName().toString());
			paramModel.setType(varElement.asType().toString());
			if (paramComments != null)
				paramModel.setComment(paramComments.remove(paramModel.getName()));

			sigModel.getParams().put(paramModel.getName(), paramModel);
		}
	}
	
	private void buildMethodThrows(ExecutableElement methodElement, String logId,
			Map<String, String> throwComments, boolean hasBpmnErrorComments, MethodSignatureFreemarkerModel sigModel) {
		for (TypeMirror thrownType : methodElement.getThrownTypes()) {
			this.logger.trace("Found thrown '{}' for '{}'", thrownType, logId);
			
			if (hasBpmnErrorComments && this.bpmnErrorSet.contains(thrownType.toString().toLowerCase()))
				continue;
			
			ThrowFreemarkerModel throwModel = new ThrowFreemarkerModel();
			throwModel.setType(thrownType.toString());
			throwModel.setComment(throwComments.remove(throwModel.getType()));
			
			sigModel.getThrowTypes().put(throwModel.getType(), throwModel);
		}
	}
	
	private void buildMethodErrors(ExecutableElement methodElement, String logId,
			Map<String, String> errorComments, MethodSignatureFreemarkerModel sigModel) {
		for (Entry<String, String> errorComment : errorComments.entrySet()) {
			this.logger.trace("Found BPMN error thrown code '{}' for '{}'", errorComment.getKey(), logId);

			ThrowFreemarkerModel errorModel = new ThrowFreemarkerModel();
			errorModel.setType(errorComment.getKey());
			errorModel.setComment(errorComment.getValue());
			
			sigModel.getBpmnErrors().put(errorModel.getType(), errorModel);
		}
	}
	
	private void buildMethodReturns(ExecutableElement methodElement, String logId, MethodSignatureFreemarkerModel sigModel) {
		this.logger.trace("Found return '{}' for '{}'", methodElement.getReturnType(), logId);
		if (!"void".equals(methodElement.getReturnType().toString()))
			sigModel.setReturnType(methodElement.getReturnType().toString());
	}
	
	private void buildMethodSignature(ExecutableElement methodElement, MethodSignatureFreemarkerModel sigModel) {
		StringBuilder methodSignature = new StringBuilder(methodElement.getSimpleName()).append("(");
		for (ParamFreemarkerModel model : sigModel.getParams().values())
			methodSignature.append(model.getName()).append(": ").append(model.getType()).append(", ");
		if (methodSignature.charAt(methodSignature.length()-1) != '(')
			methodSignature.delete(methodSignature.length()-2, methodSignature.length());
		methodSignature.append("): ").append(methodElement.getReturnType());
		
		this.logger.debug("Found method signature: {}", methodSignature);
		sigModel.setMethodSignature(methodSignature.toString());
	}
	
	private void writeIndexFile(IndexFreemarkerModel model) throws IOException, TemplateException {
		File mdfile = new File(this.outputDirectory, "index.md");
		PrintWriter mdwriter = new PrintWriter(mdfile, Charset.forName("utf-8"));
		try {
			Template template = this.fmconfig.getTemplate("index.md.ftl");
			template.process(model, mdwriter);
		} finally {
			mdwriter.close();
		}
	}
	
	private void writeBeanFile(BeanFreemarkerModel model) throws IOException, TemplateException {
		File mdfile = new File(this.outputDirectory, "bean-" + model.getBeanId() + ".md");
		PrintWriter mdwriter = new PrintWriter(mdfile, Charset.forName("utf-8"));
		try {
			Template template = this.fmconfig.getTemplate("bean.md.ftl");
			template.process(model, mdwriter);
		} finally {
			mdwriter.close();
		}
	}
	
	private String sanitize(List<? extends DocTree> docs) {
		return this.sanitize(docs.toString());
	}
	
	private String sanitize(String str) {
		for (Entry<Pattern, String> sanitize : this.sanitizations.entrySet())
			str = sanitize.getKey().matcher(str).replaceAll(sanitize.getValue());
		return str;
	}
	
	private <T> T putAdd(Map<String, List<T>> map, String key, T value) {
		List<T> values = map.get(key);
		if (values == null)
			map.put(key, values = new LinkedList<>());
		values.add(value);
		return value;
	}

}

==> v1.0.4.tar.gz: activiti-api-doclet/src/main/java/com/inteligr8/activiti/doclet/MarkdownWriter.java <==
/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.activiti.doclet;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Name;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.TypeMirror;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sun.source.doctree.DocCommentTree;
import com.sun.source.doctree.DocTree;
import com.sun.source.util.DocTrees;

import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;
import jdk.javadoc.doclet.DocletEnvironment;

/**
 * @author brian@inteligr8.com
 */
class MarkdownWriter {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final Set<String> bpmnErrorSet = new HashSet<>(Arrays.asList(
			"BpmnError".toLowerCase(),
			"org.activiti.engine.delegate.BpmnError".toLowerCase()));
	
	private final Pattern tagPattern = Pattern.compile("@([^ ]+) ?(.*)$");
	private final Pattern namedCommentPattern = Pattern.compile("([^ ]+) ?(.*)$");
	private final Map<Pattern, String> sanitizations = new FluentMap<>(new LinkedHashMap<Pattern, String>())
			.with(Pattern.compile("^[ ]+", Pattern.MULTILINE), "")
			.with(Pattern.compile("\r\n", Pattern.MULTILINE), "\n")
			.with(Pattern.compile("([^\n])\n([^\n])", Pattern.MULTILINE), "$1 $2")
			.with(Pattern.compile("([^\\.\\?!])[ ]{2}", Pattern.MULTILINE), "$1 ")
			.with(Pattern.compile("\\{ ?(@[^}]+) \\}"), "{$1}")
			.with(Pattern.compile("\\{@link ([^}]+)\\}"), "[$1]($1)")
			.with(Pattern.compile("([A-Za-z0-9\\-\\._]+@[A-Za-z0-9\\-\\._]+)"), "[$1](mailto:$1)");
	private final Pattern seeRefPattern = Pattern.compile("^([A-Za-z0-9_\\-]+)[\\.#]([A-Za-z0-9_]+)$");
	
	private final DocletEnvironment docenv;
	private final File outputDirectory;
	private final Configuration fmconfig;
	
	private String title = "API Documentation";
	private String apiName;
	
	public MarkdownWriter(DocletEnvironment docenv) throws IOException {
		this.docenv = docenv;
		this.outputDirectory = new File(".");
		this.fmconfig = this.getFreemarkerConfiguration();
	}
	
	private Configuration getFreemarkerConfiguration() throws IOException {
		Configuration fmconfig = new Configuration(Configuration.VERSION_2_3_31);
		fmconfig.setDefaultEncoding("utf-8");
		fmconfig.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);
		fmconfig.setLogTemplateExceptions(false);
		fmconfig.setWrapUncheckedExceptions(true);
		fmconfig.setFallbackOnNullLoopVariable(false);
		fmconfig.setClassLoaderForTemplateLoading(this.getClass().getClassLoader(), "/templates");
		return fmconfig;
	}
	
	public void setTitle(String title) {
		this.title = title;
	}
	
	public void setApiName(String apiName) {
		this.apiName = apiName;
	}
	
	public void write(List<ActivitiApiBeanDoc> beandocs) throws IOException, TemplateException {
		IndexFreemarkerModel model = this.buildModel(beandocs);
		
		this.outputDirectory.mkdirs();
		this.writeIndexFile(model);
		for (BeanFreemarkerModel beanModel : model.getBeans())
			this.writeBeanFile(beanModel);
	}
	
	private IndexFreemarkerModel buildModel(List<ActivitiApiBeanDoc> beandocs) throws IOException, TemplateException {
		DocTrees docs = this.docenv.getDocTrees();

		this.logger.debug("Building documentation index model");
		
		IndexFreemarkerModel indexModel = new IndexFreemarkerModel();
		indexModel.setTitle(this.title);
		indexModel.setApiName(this.apiName);

		for (ActivitiApiBeanDoc beandoc : beandocs) {
			DocCommentTree commentTree = docs.getDocCommentTree(beandoc.getClassElement());

			this.logger.debug("Building documentation bean model: {}", beandoc.getBeanId());
	
			BeanFreemarkerModel beanModel = new BeanFreemarkerModel();
			beanModel.setTitle(this.title);
			beanModel.setApiName(this.apiName);
			beanModel.setBeanId(beandoc.getBeanId());
			if (commentTree != null) {
				this.logger.debug("Found documentation for bean: {}", beandoc.getBeanId());
				
				this.buildDocumentation(commentTree, beanModel);
				for (DocTree tag : commentTree.getBlockTags()) {
					this.buildTagModel(tag, beandoc.getBeanId(),
							null, null, null, beanModel);
				}
			}
			
			if (beandoc.getDelegateMethodElement() != null) {
				beanModel.setDelegateMethod(this.buildMethodSignatureModel(beandoc, beandoc.getDelegateMethodElement()));
				indexModel.setHasDelegateBeans(true);
			}
			if (beandoc.getExecutionListenerMethodElement() != null) {
				beanModel.setExecutionListenerMethod(this.buildMethodSignatureModel(beandoc, beandoc.getExecutionListenerMethodElement()));
				indexModel.setHasDelegateBeans(true);
			}
			if (beandoc.getTaskListenerMethodElement() != null) {
				beanModel.setTaskListenerMethod(this.buildMethodSignatureModel(beandoc, beandoc.getTaskListenerMethodElement()));
				indexModel.setHasDelegateBeans(true);
			}

			Map<Name, MethodFreemarkerModel> methodOverloads = new LinkedHashMap<>();
			
			for (ExecutableElement methodElement : beandoc.getSortedMethodElements()) {
				MethodSignatureFreemarkerModel sigModel = this.buildMethodSignatureModel(beandoc, methodElement);
				
				MethodFreemarkerModel methodModel = methodOverloads.get(methodElement.getSimpleName());
				if (methodModel == null) {
					methodModel = new MethodFreemarkerModel();
					methodModel.setMethodName(methodElement.getSimpleName().toString());
					
					methodOverloads.put(methodElement.getSimpleName(), methodModel);
					beanModel.getMethods().add(methodModel);
					indexModel.setHasMethodBeans(true);
				}
				
				if (sigModel.getDocFirstSentence() != null && methodModel.getDocFirstSentence() == null)
					methodModel.setDocFirstSentence(sigModel.getDocFirstSentence());
				
				methodModel.getSignatures().add(sigModel);
			}
			
			indexModel.getBeans().add(beanModel);
		}
		
		return indexModel;
	}
	
	private MethodSignatureFreemarkerModel buildMethodSignatureModel(ActivitiApiBeanDoc beandoc, ExecutableElement methodElement) {
		String logId = beandoc.getBeanId() + "." + methodElement.toString();
		DocCommentTree methodCommentTree = this.docenv.getDocTrees().getDocCommentTree(methodElement);

		this.logger.debug("Building documentation method model: {}", logId);
		
		Map<String, String> paramComments = new HashMap<>();
		Map<String, String> throwComments = new LinkedHashMap<>();
		Map<String, String> errorComments = new LinkedHashMap<>();
		
		MethodSignatureFreemarkerModel sigModel = new MethodSignatureFreemarkerModel();
		if (methodCommentTree != null) {
			this.logger.debug("Found documentation for '{}'", logId);

			this.buildDocumentation(methodCommentTree, sigModel);
			for (DocTree tag : methodCommentTree.getBlockTags()) {
				this.buildTagModel(tag, logId,
						paramComments, throwComments, errorComments, sigModel);
			}
		}
		
		this.buildMethodArguments(methodElement, logId, paramComments, sigModel);
		this.buildMethodThrows(methodElement, logId, throwComments, !errorComments.isEmpty(), sigModel);
		this.buildMethodErrors(methodElement, logId, errorComments, sigModel);
		this.buildMethodReturns(methodElement, logId, sigModel);
		this.buildMethodBriefSignature(methodElement, sigModel);
		this.buildMethodSignature(methodElement, sigModel);

		if (!paramComments.isEmpty())
			this.logger.warn("'{}' has documented parameters {} that don't exist; ignoring ...", logId, paramComments.keySet());
		
		return sigModel;
	}
	
	private void buildDocumentation(DocCommentTree commentTree, JavadocDocumentable docuable) {
		if (commentTree.getFirstSentence() != null)
			docuable.setDocFirstSentence(this.sanitize(commentTree.getFirstSentence()));
		if (commentTree.getBody() != null)
			docuable.setDocBody(this.sanitize(commentTree.getFullBody()));
	}
	
	private void buildTagModel(DocTree tag, String logId,
			Map<String, String> paramComments,
			Map<String, String> throwComments,
			Map<String, String> errorComments,
			JavadocTaggable sigModel) {
		Matcher matcher = this.tagPattern.matcher(tag.toString());
		if (!matcher.find()) {
			this.logger.warn("A tag for '{}' is improperly formatted", logId);
			return;
		}
		
		String tagName = matcher.group(1);
		String tagComment = matcher.group(2).trim();

		this.logger.trace("Found tag '{}' for '{}'", tagName, logId);

		switch (tagName) {
			case "param":
				if (paramComments == null)
					break;
				
				matcher = this.namedCommentPattern.matcher(tagComment);
				if (matcher.find())
					paramComments.put(matcher.group(1), this.sanitize(matcher.group(2).trim()));
				else this.logger.warn("A @param for the {} bean and {} method is improperly formatted", logId);
				return;
			case "throws":
				if (throwComments == null)
					break;
				
				matcher = this.namedCommentPattern.matcher(tagComment);
				if (matcher.find()) {
					if (this.bpmnErrorSet.contains(matcher.group(1).toLowerCase())) {
						Matcher ecmatcher = this.namedCommentPattern.matcher(matcher.group(2).trim());
						if (ecmatcher.find()) {
							errorComments.put(ecmatcher.group(1), this.sanitize(ecmatcher.group(2).trim()));
						} else {
							this.logger.warn("A @throws for the {} bean and {} method is improperly formatted", logId);
						}
					} else {
						throwComments.put(matcher.group(1), this.sanitize(matcher.group(2).trim()));
					}
				}
				else this.logger.warn("A @throws for the {} bean and {} method is improperly formatted", logId);
				return;
			case "see":
				matcher = this.seeRefPattern.matcher(tagComment);
				if (matcher.find()) {
					ReferenceFreemarkerModel model = new ReferenceFreemarkerModel();
					model.setBeanId(matcher.group(1));
					model.setMethodName(matcher.group(2));
					sigModel.getSeeRefs().add(model);
					return;
				}
			default:
		}

		this.putAdd(sigModel.getTags(), tagName, this.sanitize(tagComment));
	}
	
	private void buildMethodArguments(ExecutableElement methodElement, String logId,
			Map<String, String> paramComments, MethodSignatureFreemarkerModel sigModel) {
		for (VariableElement varElement : methodElement.getParameters()) {
			this.logger.trace("Found parameter '{}' for '{}'", varElement.getSimpleName(), logId);
			
			ParamFreemarkerModel paramModel = new ParamFreemarkerModel();
			paramModel.setName(varElement.getSimpleName().toString());
			paramModel.setType(varElement.asType().toString());
			if (paramComments != null)
				paramModel.setComment(paramComments.remove(paramModel.getName()));

			sigModel.getParams().put(paramModel.getName(), paramModel);
		}
	}
	
	private void buildMethodThrows(ExecutableElement methodElement, String logId,
			Map<String, String> throwComments, boolean hasBpmnErrorComments, MethodSignatureFreemarkerModel sigModel) {
		for (TypeMirror thrownType : methodElement.getThrownTypes()) {
			this.logger.trace("Found thrown '{}' for '{}'", thrownType, logId);
			
			if (hasBpmnErrorComments && this.bpmnErrorSet.contains(thrownType.toString().toLowerCase()))
				continue;
			
			ThrowFreemarkerModel throwModel = new ThrowFreemarkerModel();
			throwModel.setType(thrownType.toString());
			throwModel.setComment(throwComments.remove(throwModel.getType()));
			
			sigModel.getThrowTypes().put(throwModel.getType(), throwModel);
		}
	}
	
	private void buildMethodErrors(ExecutableElement methodElement, String logId,
			Map<String, String> errorComments, MethodSignatureFreemarkerModel sigModel) {
		for (Entry<String, String> errorComment : errorComments.entrySet()) {
			this.logger.trace("Found BPMN error thrown code '{}' for '{}'", errorComment.getKey(), logId);

			ThrowFreemarkerModel errorModel = new ThrowFreemarkerModel();
			errorModel.setType(errorComment.getKey());
			errorModel.setComment(errorComment.getValue());
			
			sigModel.getBpmnErrors().put(errorModel.getType(), errorModel);
		}
	}
	
	private void buildMethodReturns(ExecutableElement methodElement, String logId, MethodSignatureFreemarkerModel sigModel) {
		this.logger.trace("Found return '{}' for '{}'", methodElement.getReturnType(), logId);
		if (!"void".equals(methodElement.getReturnType().toString()))
			sigModel.setReturnType(methodElement.getReturnType().toString());
	}
	
	private void buildMethodBriefSignature(ExecutableElement methodElement, MethodSignatureFreemarkerModel sigModel) {
		StringBuilder methodSignature = new StringBuilder(methodElement.getSimpleName()).append("(");
		for (ParamFreemarkerModel model : sigModel.getParams().values())
			methodSignature.append(model.getName()).append(", ");
		if (methodSignature.charAt(methodSignature.length()-1) != '(')
			methodSignature.delete(methodSignature.length()-2, methodSignature.length());
		methodSignature.append(")");
		
		this.logger.debug("Formulated method brief signature: {}", methodSignature);
		sigModel.setMethodSignature(methodSignature.toString());
	}
	
	private void buildMethodSignature(ExecutableElement methodElement, MethodSignatureFreemarkerModel sigModel) {
		StringBuilder methodSignature = new StringBuilder(methodElement.getSimpleName()).append("(");
		for (ParamFreemarkerModel model : sigModel.getParams().values())
			methodSignature.append(model.getName()).append(": ").append(model.getType()).append(", ");
		if (methodSignature.charAt(methodSignature.length()-1) != '(')
			methodSignature.delete(methodSignature.length()-2, methodSignature.length());
		methodSignature.append("): ").append(methodElement.getReturnType());
		
		this.logger.debug("Formulated method signature: {}", methodSignature);
		sigModel.setMethodSignature(methodSignature.toString());
	}
	
	private void writeIndexFile(IndexFreemarkerModel model) throws IOException, TemplateException {
		File mdfile = new File(this.outputDirectory, "index.md");
		PrintWriter mdwriter = new PrintWriter(mdfile, Charset.forName("utf-8"));
		try {
			Template template = this.fmconfig.getTemplate("index.md.ftl");
			template.process(model, mdwriter);
		} finally {
			mdwriter.close();
		}
	}
	
	private void writeBeanFile(BeanFreemarkerModel model) throws IOException, TemplateException {
		File mdfile = new File(this.outputDirectory, "bean-" + model.getBeanId() + ".md");
		PrintWriter mdwriter = new PrintWriter(mdfile, Charset.forName("utf-8"));
		try {
			Template template = this.fmconfig.getTemplate("bean.md.ftl");
			template.process(model, mdwriter);
		} finally {
			mdwriter.close();
		}
	}
	
	private String sanitize(List<? extends DocTree> docs) {
		return this.sanitize(docs.toString());
	}
	
	private String sanitize(String str) {
		for (Entry<Pattern, String> sanitize : this.sanitizations.entrySet())
			str = sanitize.getKey().matcher(str).replaceAll(sanitize.getValue());
		return str;
	}
	
	private <T> T putAdd(Map<String, List<T>> map, String key, T value) {
		List<T> values = map.get(key);
		if (values == null)
			map.put(key, values = new LinkedList<>());
		values.add(value);
		return value;
	}

}
